import asyncio
import json
from asyncio import gather
from contextlib import asynccontextmanager
from logging import getLogger
from typing import Never

try:
    from jsonschema_pydantic import jsonschema_to_pydantic
    from langchain_core.tools import BaseTool
    from mcp import ClientSession, StdioServerParameters
    from mcp.client.stdio import stdio_client
    from mcp.types import ListToolsResult, Tool
    from pydantic import BaseModel

except ImportError as e:
    raise ImportError(
        f"Please install the required packages by running: {e}"
        "pip install langchain-core mcp jsonschema-pydantic"
    ) from e


from .data_models.mcp_servers import McpServers

log = getLogger("mcp_services.LangChainMcp")


class ConfigurationError(Exception):
    """Error raised when there is an issue with the configuration."""

    pass


class LangChainMcp:
    """List all the available tools for all servers."""

    def __init__(self, config_path: str = "./claude_mcp_config.json"):
        self.config_path = config_path
        self._server_configs: McpServers | None = None

    @property
    def server_configs(self) -> McpServers:
        """Get the server configurations."""
        if self._server_configs is None:
            self._server_configs = self._load_config()
        return self._server_configs

    def _load_config(self) -> McpServers:
        """Load mcp server configurations from file."""
        if self._server_configs is None:
            try:
                with open(self.config_path) as f:
                    config_data = json.load(f)
                self._server_configs = McpServers(**config_data)
            except FileNotFoundError as error:
                raise ConfigurationError(
                    f"Configuration file not found: {self.config_path}"
                ) from error
            except json.JSONDecodeError as error:
                raise ConfigurationError(
                    f"Invalid JSON in configuration file: {self.config_path}"
                ) from error
        return self._server_configs

    @staticmethod
    @asynccontextmanager
    async def get_session(server_params: StdioServerParameters) -> ClientSession:
        """Get a client MCP session.

        Args:
            server_params (StdioServerParameters): The server parameters.

        Yields:
            ClientSession: The client session for the MCP server.

        """
        async with (
            stdio_client(server_params) as (read, write),
            ClientSession(read, write) as session,
        ):
            await session.initialize()
            yield session

    async def _fetch_langchain_tools(
        self, server_name: str, params: StdioServerParameters
    ) -> list[BaseTool]:
        """List the available tools to call.

        Args:
            server_name (Server): The name of the server to create a session.
            params (StdioServerParameters): The server parameters.

        Returns:
            list[BaseTool]: The available tools for the server as LangChain tools.

        """
        log.info(f"Listing tools for server: {server_name}")
        async with self.get_session(params) as session:
            list_tool_results: ListToolsResult = await session.list_tools()
            all_tools: list[BaseTool] = []
            for tool in list_tool_results.tools:
                adaptor = await self.convert_mcp_to_langchain_tools(
                    server_name, tool, params
                )
                all_tools.append(adaptor)
            return all_tools

    @staticmethod
    async def convert_mcp_to_langchain_tools(
        server_name: str, tool: Tool, params: StdioServerParameters
    ) -> BaseTool:
        """Convert an MCP tool to a LangChain tool.

        Args:
            server_name (str): The name of the mcp server.
            tool (Tool): The MCP tool to convert to LangChain Tool.
            params (StdioServerParameters): The server parameters.
        """
        logger = getLogger(f"mcp_services.{server_name}.{tool.name}")

        class McpLangchainAdaptor(BaseTool):
            """Adaptor class to convert MCP tool to LangChain tool."""

            name: str = tool.name
            description: str = tool.description
            args_schema: type[BaseModel] = jsonschema_to_pydantic(tool.inputSchema)

            @staticmethod
            @asynccontextmanager
            async def get_session(
                server_params: StdioServerParameters,
            ) -> ClientSession:
                """Get a client MCP session.

                Args:
                    server_params (StdioServerParameters): The server parameters.

                Yields:
                    ClientSession: The client session for the MCP server.

                """
                async with (
                    stdio_client(server_params) as (read, write),
                    ClientSession(read, write) as session,
                ):
                    await session.initialize()
                    logger.info(f"Session created for {server_name}:{tool.name}")
                    yield session

            def _run(self, **kwargs: dict) -> Never:
                raise NotImplementedError("Only async operation is supported")

            async def _arun(self, **kwargs: dict) -> list:
                async with self.get_session(params) as session:
                    logger.info(
                        f"Calling tool {server_name}:{tool.name} "
                        f"with arguments {kwargs}"
                    )
                    results = await session.call_tool(self.name, kwargs)
                    return results

        return McpLangchainAdaptor()

    async def fetch_all_server_tools(self) -> list[BaseTool]:
        """List the available tools for all servers.

        Returns:
            list[BaseTool]: The available tools for all servers as LangChain tools.
        """
        coroutines = [
            self._fetch_langchain_tools(server_name, params)
            for server_name, params in self.server_configs.mcpServers.items()
        ]
        results = await gather(*coroutines)
        return [item for sublist in results for item in sublist]

    def list_mcp_tools(self) -> list[BaseTool]:
        """Make async call to list all the available tools for all mcp servers."""
        return asyncio.run(self.fetch_all_server_tools())



import os
import sys

from mcp import StdioServerParameters
from pydantic import BaseModel, Field, field_validator

# Environment variables to inherit by default
DEFAULT_INHERITED_ENV_VARS = (
    [
        "APPDATA",
        "HOMEDRIVE",
        "HOMEPATH",
        "LOCALAPPDATA",
        "PATH",
        "PROCESSOR_ARCHITECTURE",
        "SYSTEMDRIVE",
        "SYSTEMROOT",
        "TEMP",
        "USERNAME",
        "USERPROFILE",
    ]
    if sys.platform == "win32"
    else ["HOME", "LOGNAME", "PATH", "SHELL", "TERM", "USER"]
)


def get_default_environment() -> dict[str, str]:
    """Returns a default environment.

    Inherited object are environment variables deemed safe to inherit.
    """
    env: dict[str, str] = {}

    for key in DEFAULT_INHERITED_ENV_VARS:
        value = os.environ.get(key)
        if value is None:
            continue

        if value.startswith("()"):
            # Skip functions, which are a security risk
            continue

        env[key] = value

    return env


class StdioServerParameters(StdioServerParameters):
    """Data model for the stdio server parameters."""

    @field_validator("env", mode="before")
    @classmethod
    def parse_env(cls, env: dict) -> dict[str, str] | None:
        """Parse the environment variables.

        For each environment variable that starts with "ENV_", replace the value with
        the value of the corresponding environment variable.

        Args:
            env: The environment variables.
        """
        default_env = get_default_environment()
        for key in env:
            if env[key].startswith("ENV_"):
                env_var = os.environ.get(env[key][4:])
                if env_var is None:
                    raise ValueError(f"Environment variable {env[key][4:]} not found.")
                env[key] = env_var
        return default_env | env


class McpServers(BaseModel):
    """Data model for mcp servers."""

    mcpServers: dict[str, StdioServerParameters] = Field(
        ...,
        description="The list of mcp servers configurations.",
    )

    @field_validator("mcpServers", mode="before")
    @classmethod
    def parse_configs(cls, mcpServers: dict) -> dict[any, StdioServerParameters]:
        """Parse the mcp servers configurations.

        Args:
            mcpServers: The mcp servers configurations.

        Returns:
            list[StdioServerParameters]: The list of mcp servers configurations.
        """
        return {
            server: StdioServerParameters(**mcpServers[server]) for server in mcpServers
        }



# /// script
# dependencies = [
#   "langchain>=0.3.9",
#   "langgraph>=0.2.53",
#   "langchain-openai>=0.2.10",
#   "langchain-community>=0.3.9",
#   "langchain-mcp-connect>=2.0.0",
# ]
# ///

import asyncio
import logging
import os

from dotenv import load_dotenv
from langchain_core.messages import HumanMessage
from langchain_openai import ChatOpenAI
from langgraph.prebuilt import create_react_agent

from langchain_mcp_connect import LangChainMcp

load_dotenv()

logging.basicConfig(level=logging.INFO)
log = logging.getLogger("LangChainMcp")

if "GITHUB_PERSONAL_ACCESS_TOKEN" not in os.environ:
    raise ValueError(
        "Please set the GITHUB_PERSONAL_ACCESS_TOKEN environment variable."
    )
if "OPENAI_API_KEY" not in os.environ:
    raise ValueError("Please set the OPENAI_API_KEY environment variable.")


if __name__ == "__main__":

    QUERY = "What tools do you have access to?"

    # Define the llm
    llm = ChatOpenAI(
        model="gpt-4o",
        model_kwargs={
            "max_tokens": 4096,
            "temperature": 0.0,
        },
    )

    # Fetch tools
    mcp = LangChainMcp()
    tools = mcp.list_mcp_tools()

    # Bind tools to the agent
    agent_executor = create_react_agent(llm, tools)
    human_message = dict(messages=[HumanMessage(content=QUERY)])

    # Run the agent asynchronously
    response = asyncio.run(agent_executor.ainvoke(input=human_message))

    log.info(response)



import os

import pytest
from langchain_core.tools import BaseTool as LangChainBaseTool
from mcp.types import Tool as McpTool
from pydantic import BaseModel

from langchain_mcp_connect import LangChainMcp


@pytest.fixture
def mock_mcp_tool() -> McpTool:
    """Return a mock McpTool object."""
    return McpTool(
        name="get_file_contents",
        description="Retrieve the file contents of a file from a github repository",
        inputSchema={
            "type": "object",
            "properties": {
                "owner": {
                    "type": "string",
                    "description": "Repository owner (username or organization)",
                },
                "repo": {"type": "string", "description": "Repository name"},
                "path": {
                    "type": "string",
                    "description": "Path to the file or directory",
                },
                "branch": {
                    "type": "string",
                    "description": "Branch to get contents from",
                },
            },
            "required": ["owner", "repo", "path"],
            "additionalProperties": False,
            "$schema": "http://json-schema.org/draft-07/schema#",
        },
    )


class TestLangChainMCP:
    """Test the LangChainMcp class."""

    def test_init(self) -> None:
        """Test the __init__ method.

        Assets that the config_path is set to "./claude_mcp_config.json" and
        the _server_configs is None.
        """
        lcp = LangChainMcp()
        assert lcp.config_path == "./claude_mcp_config.json"
        assert lcp._server_configs is None

    def test_server_configs(self) -> None:
        """Test the server_configs property.

        Assets that the server_configs property returns the _load_config method.
        """
        os.environ["GITHUB_PERSONAL_ACCESS_TOKEN"] = "test"
        os.environ["OPENAI_API_KEY"] = "test"

        lcp = LangChainMcp()

        assert [key for key in lcp.server_configs.mcpServers] == [
            "git",
            "filesystem",
            "github",
        ]
        assert lcp.server_configs == lcp._load_config()

    @pytest.mark.asyncio
    async def test_convert_mcp_to_langchain_tools(self, mock_mcp_tool: McpTool) -> None:
        """Test the convert_mcp_to_langchain_tools method.

        Assets that the _convert_mcp_to_langchain_tools method returns a list of
        BaseTool objects.
        """
        lcp = LangChainMcp()
        param = lcp.server_configs.mcpServers["github"]
        langchain_tool = await lcp.convert_mcp_to_langchain_tools(
            "github", mock_mcp_tool, param
        )

        assert isinstance(langchain_tool, LangChainBaseTool)
        assert langchain_tool.name == "get_file_contents"
        assert langchain_tool.description == (
            "Retrieve the file contents of a file " "from a github repository"
        )
        assert issubclass(langchain_tool.args_schema, BaseModel)

    @pytest.mark.asyncio
    async def test_mcp_langchain_adaptor_methods(self, mock_mcp_tool: McpTool) -> None:
        """Test that the correct methods exists for McpLangChainAdaptor.

        Asserts that only async operations are valid
        """
        lcp = LangChainMcp()

        param = lcp.server_configs.mcpServers["github"]
        langchain_tool = await lcp.convert_mcp_to_langchain_tools(
            "github", mock_mcp_tool, param
        )

        assert hasattr(langchain_tool, "get_session")
        assert hasattr(langchain_tool, "_arun")
        assert hasattr(langchain_tool, "_run")

        with pytest.raises(NotImplementedError):
            langchain_tool._run()



import os

import pytest

from langchain_mcp_connect.data_models import StdioServerParameters


class TestStdioServerParameters:
    """Test the StdioServerParameters model."""

    def test_environment_variables(self) -> None:
        """Test that the model sources environment variables with ENV_ prefix.

        Asserts that the environment variable is correctly sourced when ENV_ prefix
        is used.
        """
        os.environ["TEST_ENV_VAR"] = "wubba.lubba.dub.dub"
        stdio_server_parameters = StdioServerParameters(
            command="uvx", args=["./"], env={"TEST_ENV_VAR": "ENV_TEST_ENV_VAR"}
        )
        assert isinstance(stdio_server_parameters, StdioServerParameters)
        assert stdio_server_parameters.command == "uvx"
        assert stdio_server_parameters.args == ["./"]
        assert stdio_server_parameters.env["TEST_ENV_VAR"] == "wubba.lubba.dub.dub"

    def test_environment_variables_raise(self) -> None:
        """Test that the model sources environment variables with ENV_ prefix.

        Asserts that the environment variable is correctly sourced when ENV_ prefix
        is used.
        """
        with pytest.raises(ValueError):
            StdioServerParameters(
                command="uvx", args=["./"], env={"ANOTHER_VAR": "ENV_ANOTHER_VAR"}
            )
